package server;

import java.util.*;
import java.net.*;
import java.io.*;

/**
 * The central server, provided by the exercise assistants, that
 * accepts user registrations and also gives out a list with all
 * registered users. A registration is guaranteed to last only {@link
 * #LIFETIME} seconds long, after that it expires (sooner or later) if
 * the registration is not renewed. A user is represented by his IP
 * address, port number and self-chosen name.
 */
public class UserIDServer implements ProtocolConstants, Runnable {

	/** The actual socket that listens on the port. */
	private ServerSocket server;

	/** How long a registration is valid, in seconds. */
	public final int LIFETIME;

	/** Maps id's to User's */
	HashMap map = new HashMap();

	/** Administrative thread issues. */
	private ThreadPool pool;

	public static boolean DEBUG = false;

	/* logging */
	public static FileWriter fileWriter;

	/**
	 * Creates an ID server with the specified lifetime of registrations.
	 *
	 * @throw a {@link java.io.IOException} if it failed to properly
	 * set up everything
	 */
	public UserIDServer(int lifetime) throws IOException {
		LIFETIME = lifetime;
		TimerTask task = new MapCleaner();
		Timer timer = new Timer();
		timer.schedule(task, LIFETIME * 1000, LIFETIME * 500);

		SocketWorker worker = new SocketWorker(this);
		pool = new ThreadPool(worker, 10, 100);

		server = new ServerSocket(PORT);

		fileWriter = new FileWriter("users.log");
	}

	private Long convert(byte[] addr, int port) {
		long tmp = (long) (((addr[0] & 0xff) << 40) | ((addr[1] & 0xff) << 32) | ((addr[2] & 0xff) << 24) | ((addr[3] & 0xff) << 16)) + port;
		return new Long(tmp);
	}

	/**
	 * @return If we already know a user with the given ID, then
	 * return him, otherwise <code>null</code>
	 */
	User getUser(byte[] addr, int port) {
		return (User)map.get(convert(addr, port));
	}

	/**
	 * @return all users currently in the list
	 */
	User[] getUsers() {
		return (User[])map.values().toArray(new User[map.size()]);
	}

	/**
	 * Registers a user and also remembers the time of the registration.
	 */
	synchronized void registerUser(User user) {
		user.setTimestamp();
		map.put(convert(user.getAddress(), user.getPort()), user);
	}

	/**
	 * A {@link java.util.TimerTask} that runs in the background
	 * periodically and cleans up the timed out users.
	 */
	private class MapCleaner extends TimerTask {
		public void run() {
			/* that one is synchronized on UserIDServer */
			cleanMap();
		}
	}

	synchronized void cleanMap() {
		long old = LIFETIME * 1000;
		long time = System.currentTimeMillis();
		for (Iterator iter = map.entrySet().iterator(); iter.hasNext();) {
			Map.Entry entry = (Map.Entry)iter.next();
			User u = (User)entry.getValue();
			if (time - u.getTimestamp() > old) {
				iter.remove();

				try {
					fileWriter.write("--- ");
					fileWriter.write((new java.util.Date(System.currentTimeMillis())).toString());
					fileWriter.write(" *** ");
					fileWriter.write(u.toString() + "\n");
					fileWriter.flush();
				} catch (IOException ioe) {
					ioe.printStackTrace();
				}
			}
		}
	}

	/** Start the actual server. */
	public void run() {
		while (true) {
			try {
				Socket client = server.accept();
				ReusableThread worker = pool.getThread();
				Object[] args = { client };
				worker.setReady(args);
			} catch (IOException ioe) {
				if (DEBUG) {
					System.err.println("Could not accept connection.");
					ioe.printStackTrace();
				}
			}
		}
	}

	/** Clean-up of sockets. */
	protected void finalize() throws IOException {
		server.close();
	}

	/*------------------------------------*/

	public static void main(String[] args) {
		if (args.length < 1) {
			usage();
			System.exit(0);
		}

		UserIDServer theServer = null;
		try {
			theServer = new UserIDServer(Integer.parseInt(args[0]));
		} catch (IOException ioe) {
			System.err.println("Could not start server. Exiting.");
			System.exit(-1);
		}
		if ((args.length > 1) && args[1].startsWith("-d"))
			DEBUG = true;

		System.err.println("Starting server... ");
		//Thread thread = new Thread(server, "server");
		theServer.run();
	}

	private static void usage() {
		System.out.println("Parameters: <timeout> [-d]");
		System.out.println("where\t<timeout> = (min) time that a " + "registration is valid (in seconds)\n" + "\t-d = print debugging info");
	}

}
